#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include "mt64.h"
#include "percolating_clusters.h"
#include "new_theory.h"


#define SENS 1e-7




void add_edge_matrix (int i, int n, int **bond)
{
  bond[0][i] ++;
  bond[i] = (int *)realloc(bond[i],(bond[0][i]+1)*sizeof(int));
  bond[i][bond[0][i]] = n;
}


int find_edge_matrix (int q, int n, int **bond, int **edges)
{

  int k, s, t, v;
  for(k=1;k<=bond[0][q];k++)
    {
      s = bond[q][k];
      t = edges[0][s];
      v = edges[1][s];
      if (v==n) return s;
    }
  printf("#Error\n");
  return -1;
  
}





void find_transition_point_new_theory (int N, int **onlyA, int **onlyB, int **intersection, double *res)
{

  int i, j, n, m;


  int MA = 0;
  for(i=1;i<=N;i++) MA+=onlyA[0][i];
  int MB = 0;
  for(i=1;i<=N;i++) MB+=onlyB[0][i];
  int MI = 0;
  for(i=1;i<=N;i++) MI+=intersection[0][i];

  int M = MI + MA + MB;
  double **prob_zeta = (double **)malloc((M+1)*sizeof(double *));
  for(i=1;i<=M;i++) prob_zeta[i] = (double *)malloc(3*sizeof(double));
  int **edges = (int **)malloc(4*sizeof(int *));
  edges[0] = (int *)malloc((M+1)*sizeof(int));
  edges[1] = (int *)malloc((M+1)*sizeof(int));
  edges[2] = (int *)malloc((M+1)*sizeof(int));
  edges[3] = (int *)malloc((M+1)*sizeof(int));

  int **type_bond = (int **)malloc((N+1)*sizeof(int *));
  int **bond_to_edge = (int **)malloc((N+1)*sizeof(int *));
  int **rev_bond = (int **)malloc((N+1)*sizeof(int *));
  rev_bond[0] = (int *)malloc((N+1)*sizeof(int));
  for(i=1;i<=N;i++)
    {
      rev_bond[0][i] = onlyA[0][i]+onlyB[0][i]+intersection[0][i];
      rev_bond[i] = (int *)malloc((rev_bond[0][i]+1)*sizeof(int));
      bond_to_edge[i] = (int *)malloc((rev_bond[0][i]+1)*sizeof(int));
      type_bond[i] = (int *)malloc((rev_bond[0][i]+1)*sizeof(int));
      rev_bond[0][i] = 0;
    }

  int **matrix = (int **)malloc((M+1)*sizeof(int *));
  matrix[0] = (int *)malloc((M+1)*sizeof(int));
  for(i=1;i<=M;i++)
    {
      matrix[i] = (int *)malloc(1*sizeof(int));
      matrix[0][i] = 0;
    }



  double *probA = (double *)malloc((MA+1)*sizeof(double));
  double *probB = (double *)malloc((MB+1)*sizeof(double));
  double **probI = (double **)malloc((MI+1)*sizeof(double *));
  for(i=1;i<=MI;i++) probI[i] = (double *)malloc(3*sizeof(double));




  M = 0;

  MA = 0;
  for(i=1;i<=N;i++) 
    {
      for(j=1;j<=onlyA[0][i];j++)
	{
	  n = onlyA[i][j];
	 
	  MA ++;

	  M++;


	  edges[0][M] = i;
	  edges[1][M] = n;
	  edges[2][M] = 1;
	  edges[3][M] = MA;

	  rev_bond[0][i]++;
	  rev_bond[i][rev_bond[0][i]] = M;
	  type_bond[i][rev_bond[0][i]] = 1;
	  bond_to_edge[i][rev_bond[0][i]] = MA;
	  

	}
    }

  
  MB = 0;
  for(i=1;i<=N;i++) 
    {
      for(j=1;j<=onlyB[0][i];j++)
	{
	  n = onlyB[i][j];
	 
	  MB ++;
	  
	  M++;


	  edges[0][M] = i;
	  edges[1][M] = n;
	  edges[2][M] = 2;
	  edges[3][M] = MB;

	  rev_bond[0][i]++;
	  rev_bond[i][rev_bond[0][i]] = M;
	  type_bond[i][rev_bond[0][i]] = 2;
	  bond_to_edge[i][rev_bond[0][i]] = MB;
	  
	}
    }

  MI = 0;
  for(i=1;i<=N;i++) 
    {
      for(j=1;j<=intersection[0][i];j++)
	{
	  n = intersection[i][j];

	  MI ++;
	  
	  M++;


	  edges[0][M] = i;
	  edges[1][M] = n;
	  edges[2][M] = 0;
	  edges[3][M] = MI;

	  rev_bond[0][i]++;
	  rev_bond[i][rev_bond[0][i]] = M;
	  type_bond[i][rev_bond[0][i]] = 0;
	  bond_to_edge[i][rev_bond[0][i]] = MI;
	  
	}
    }





  int q, t;
  
  for(i=1;i<=M;i++)
    {

      n = edges[0][i];
      m = edges[1][i];

      for(j=1;j<=intersection[0][n];j++)
	{

	  q = intersection[n][j];
	  if (q != m)
	    {
	      //find edge index q -> n
	      t = find_edge_matrix (q, n, rev_bond, edges);
	      if (t>0) 
		{
		  add_edge_matrix (i, t, matrix);
		  //printf("I: (%d, %d) --> (%d, %d)\n",n,m,edges[0][t], edges[1][t]);
		} 
	    }

	}

      for(j=1;j<=onlyA[0][n];j++)
	{

	  q = onlyA[n][j];
	  if (q != m)
	    {
	      //find edge index q -> n
	      t = find_edge_matrix (q, n, rev_bond, edges);
	      if (t>0) 
		{
		  add_edge_matrix (i, t, matrix);
		  //printf("A: (%d, %d) --> (%d, %d)\n",n,m,edges[0][t], edges[1][t]);
		} 
	    }

	}

      for(j=1;j<=onlyB[0][n];j++)
	{

	  q = onlyB[n][j];
	  if (q != m)
	    {
	      //find edge index q -> n
	      t = find_edge_matrix (q, n, rev_bond, edges);
	      if (t>0) 
		{
		  add_edge_matrix (i, t, matrix);
		  //printf("B: (%d, %d) --> (%d, %d)\n",n,m,edges[0][t], edges[1][t]);
		} 
	    }

	}


      //printf("%d %d\n",i,matrix[0][i]);

    }


  

  


  

  double err;
  double order, order1, order2;

  double dp = 1e-4;
  double p2  = 1.0-dp;
  double p1 = dp;
  double p;





  while(p2-p1>dp)
    {
  
      //p2
      /*for(i=1;i<=MA;i++) probA[i] = 1.0;
      for(i=1;i<=MB;i++) probB[i] = 1.0;
      for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = 1.0;*/
      for(i=1;i<=MA;i++) probA[i] = genrand64_real3();
      for(i=1;i<=MB;i++) probB[i] = genrand64_real3();
      for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = genrand64_real3();
      err = 1.0;
      while(err>SENS)
	{
	  err =  single_iteration_novel (M, matrix, edges, prob_zeta, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p2);
	}
      order2 = compute_order (prob_zeta, M, edges, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p2);
      
      //p1
     /*for(i=1;i<=MA;i++) probA[i] = 1.0;
      for(i=1;i<=MB;i++) probB[i] = 1.0;
      for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = 1.0;*/
      for(i=1;i<=MA;i++) probA[i] = genrand64_real3();
      for(i=1;i<=MB;i++) probB[i] = genrand64_real3();
      for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = genrand64_real3();
      err = 1.0;
      while(err>SENS)
	{
	  err =  single_iteration_novel (M, matrix, edges, prob_zeta, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p1);
	}
      order1 = compute_order (prob_zeta, M, edges, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p1);


      //p
      p = (p1 + p2) /2.0;
     /*for(i=1;i<=MA;i++) probA[i] = 1.0;
      for(i=1;i<=MB;i++) probB[i] = 1.0;
      for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = 1.0;*/
      for(i=1;i<=MA;i++) probA[i] = genrand64_real3();
      for(i=1;i<=MB;i++) probB[i] = genrand64_real3();
      for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = genrand64_real3();
      err = 1.0;
      while(err>SENS)
	{
	  err =  single_iteration_novel (M, matrix, edges, prob_zeta, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p);
	}
      order = compute_order (prob_zeta, M, edges, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p);



      printf("#%g %g\t%g %g\t%g %g\n",p1,order1,p2,order2,p,order);


      if(order1>1.0/(double)N && order2>1.0/(double)N)
	{
	  printf("#Error A\n");
	}

      if(order1<1.0/(double)N && order2<1.0/(double)N)
	{
	  printf("#Error B\n");
	}


      if(order1 < 1.0/(double)N && order < 1.0 /(double)N)
	{
	  p1 = p;
	}

      if(order2 > 1.0/(double)N && order > 1.0 /(double)N)
	{
	  p2 = p;
	}

    }



  res[0] = order2 - order1;
  res[1] = p;




  ////////////////////
  
  for(i=0;i<=M;i++) free(matrix[i]);
  free(matrix);

  free(probA);
  free(probB);
  for(i=1;i<=MI;i++) free(probI[i]);
  free(probI);
  


  for(i=1;i<=M;i++) free(prob_zeta[i]);
  free(prob_zeta);
  free(edges[0]);
  free(edges[1]);
  free(edges[2]);
  free(edges[3]);
  free(edges);

  free(rev_bond[0]);
  for(i=1;i<=N;i++)
    {
      free(rev_bond[i]);
      free(type_bond[i]);
      free(bond_to_edge[i]);
    }
  free(rev_bond);
  free(type_bond);
  free(bond_to_edge);
}





void order_parameter_new_theory (int N, int **onlyA, int **onlyB, int **intersection, char *filename)
{

  int i, j, n, m;


  int MA = 0;
  for(i=1;i<=N;i++) MA+=onlyA[0][i];
  int MB = 0;
  for(i=1;i<=N;i++) MB+=onlyB[0][i];
  int MI = 0;
  for(i=1;i<=N;i++) MI+=intersection[0][i];

  int M = MI + MA + MB;
  double **prob_zeta = (double **)malloc((M+1)*sizeof(double *));
  for(i=1;i<=M;i++) prob_zeta[i] = (double *)malloc(3*sizeof(double));
  int **edges = (int **)malloc(4*sizeof(int *));
  edges[0] = (int *)malloc((M+1)*sizeof(int));
  edges[1] = (int *)malloc((M+1)*sizeof(int));
  edges[2] = (int *)malloc((M+1)*sizeof(int));
  edges[3] = (int *)malloc((M+1)*sizeof(int));

  int **type_bond = (int **)malloc((N+1)*sizeof(int *));
  int **bond_to_edge = (int **)malloc((N+1)*sizeof(int *));
  int **rev_bond = (int **)malloc((N+1)*sizeof(int *));
  rev_bond[0] = (int *)malloc((N+1)*sizeof(int));
  for(i=1;i<=N;i++)
    {
      rev_bond[0][i] = onlyA[0][i]+onlyB[0][i]+intersection[0][i];
      rev_bond[i] = (int *)malloc((rev_bond[0][i]+1)*sizeof(int));
      bond_to_edge[i] = (int *)malloc((rev_bond[0][i]+1)*sizeof(int));
      type_bond[i] = (int *)malloc((rev_bond[0][i]+1)*sizeof(int));
      rev_bond[0][i] = 0;
    }

  int **matrix = (int **)malloc((M+1)*sizeof(int *));
  matrix[0] = (int *)malloc((M+1)*sizeof(int));
  for(i=1;i<=M;i++)
    {
      matrix[i] = (int *)malloc(1*sizeof(int));
      matrix[0][i] = 0;
    }



  double *probA = (double *)malloc((MA+1)*sizeof(double));
  double *probB = (double *)malloc((MB+1)*sizeof(double));
  double **probI = (double **)malloc((MI+1)*sizeof(double *));
  for(i=1;i<=MI;i++) probI[i] = (double *)malloc(3*sizeof(double));




  M = 0;

  MA = 0;
  for(i=1;i<=N;i++) 
    {
      for(j=1;j<=onlyA[0][i];j++)
	{
	  n = onlyA[i][j];
	 
	  MA ++;

	  M++;


	  edges[0][M] = i;
	  edges[1][M] = n;
	  edges[2][M] = 1;
	  edges[3][M] = MA;

	  rev_bond[0][i]++;
	  rev_bond[i][rev_bond[0][i]] = M;
	  type_bond[i][rev_bond[0][i]] = 1;
	  bond_to_edge[i][rev_bond[0][i]] = MA;
	  

	}
    }

  
  MB = 0;
  for(i=1;i<=N;i++) 
    {
      for(j=1;j<=onlyB[0][i];j++)
	{
	  n = onlyB[i][j];
	 
	  MB ++;
	  
	  M++;


	  edges[0][M] = i;
	  edges[1][M] = n;
	  edges[2][M] = 2;
	  edges[3][M] = MB;

	  rev_bond[0][i]++;
	  rev_bond[i][rev_bond[0][i]] = M;
	  type_bond[i][rev_bond[0][i]] = 2;
	  bond_to_edge[i][rev_bond[0][i]] = MB;
	  
	}
    }

  MI = 0;
  for(i=1;i<=N;i++) 
    {
      for(j=1;j<=intersection[0][i];j++)
	{
	  n = intersection[i][j];

	  MI ++;
	  
	  M++;


	  edges[0][M] = i;
	  edges[1][M] = n;
	  edges[2][M] = 0;
	  edges[3][M] = MI;

	  rev_bond[0][i]++;
	  rev_bond[i][rev_bond[0][i]] = M;
	  type_bond[i][rev_bond[0][i]] = 0;
	  bond_to_edge[i][rev_bond[0][i]] = MI;
	  
	}
    }





  int q, t;
  
  for(i=1;i<=M;i++)
    {

      n = edges[0][i];
      m = edges[1][i];

      for(j=1;j<=intersection[0][n];j++)
	{

	  q = intersection[n][j];
	  if (q != m)
	    {
	      //find edge index q -> n
	      t = find_edge_matrix (q, n, rev_bond, edges);
	      if (t>0) 
		{
		  add_edge_matrix (i, t, matrix);
		  //printf("I: (%d, %d) --> (%d, %d)\n",n,m,edges[0][t], edges[1][t]);
		} 
	    }

	}

      for(j=1;j<=onlyA[0][n];j++)
	{

	  q = onlyA[n][j];
	  if (q != m)
	    {
	      //find edge index q -> n
	      t = find_edge_matrix (q, n, rev_bond, edges);
	      if (t>0) 
		{
		  add_edge_matrix (i, t, matrix);
		  //printf("A: (%d, %d) --> (%d, %d)\n",n,m,edges[0][t], edges[1][t]);
		} 
	    }

	}

      for(j=1;j<=onlyB[0][n];j++)
	{

	  q = onlyB[n][j];
	  if (q != m)
	    {
	      //find edge index q -> n
	      t = find_edge_matrix (q, n, rev_bond, edges);
	      if (t>0) 
		{
		  add_edge_matrix (i, t, matrix);
		  //printf("B: (%d, %d) --> (%d, %d)\n",n,m,edges[0][t], edges[1][t]);
		} 
	    }

	}


      //printf("%d %d\n",i,matrix[0][i]);

    }


  

  



  FILE *f;
  f = fopen(filename,"w");
  

  double err;
  double order;


  double dp = 1e-3;
  double p  = 1.0 - dp;

  /*for(i=1;i<=MA;i++) probA[i] = 1.0;
    for(i=1;i<=MB;i++) probB[i] = 1.0;
    for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = 1.0;*/
  for(i=1;i<=MA;i++) probA[i] = genrand64_real3();
  for(i=1;i<=MB;i++) probB[i] = genrand64_real3();
  for(i=1;i<=MI;i++) probI[i][0] = probI[i][1] = probI[i][2] = genrand64_real3();



  while(p >= dp)
    {

      err = 1.0;

      while(err>SENS)
	{

	  err =  single_iteration_novel (M, matrix, edges, prob_zeta, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p);
	  //printf("%g %g\n",p,err); fflush(stdout);
	}

      //printf("--> %g %g\n",p,err);


      order = compute_order (prob_zeta, M, edges, N, rev_bond, type_bond, bond_to_edge, probI, probA, probB, p);


      printf("--> %g %g\n",p,order);

      fprintf(f,"%g %g\n",p,order);

      //exit(0);

      p -= dp;

    }


  fclose(f);




  ////////////////////
  
  for(i=0;i<=M;i++) free(matrix[i]);
  free(matrix);

  free(probA);
  free(probB);
  for(i=1;i<=MI;i++) free(probI[i]);
  free(probI);
  


  for(i=1;i<=M;i++) free(prob_zeta[i]);
  free(prob_zeta);
  free(edges[0]);
  free(edges[1]);
  free(edges[2]);
  free(edges[3]);
  free(edges);

  free(rev_bond[0]);
  for(i=1;i<=N;i++)
    {
      free(rev_bond[i]);
      free(type_bond[i]);
      free(bond_to_edge[i]);
    }
  free(rev_bond);
  free(type_bond);
  free(bond_to_edge);
}




void compute_zeta (double **prob_zeta, int N, int **rev_bond, int **type_bond, int **bond_to_edge, double **probI, double *probA, double *probB)
{

  int s, t, n, i, j;
  
  for(i=1;i<=N;i++)
    {

      for(j=1;j<=rev_bond[0][i];j++)
	{

	  n = rev_bond[i][j];
	  t = type_bond[i][j];
	  s = bond_to_edge[i][j];

	  if(t == 0)
	    {
	      prob_zeta[n][0] = probI[s][0] + probI[s][1] + probI[s][2];
	      prob_zeta[n][1] = probI[s][0] + probI[s][1];
	      prob_zeta[n][2] = probI[s][0] + probI[s][2];
	    }

	  if (t == 1){
	    prob_zeta[n][0] = probA[s];
	    prob_zeta[n][1] = probA[s];
	    prob_zeta[n][2] = 0.0;
	  }

	  if (t == 2){
	    prob_zeta[n][0] = probB[s];
	    prob_zeta[n][1] = 0.0;
	    prob_zeta[n][2] = probB[s];
	  }
	}
    }
}



void compute_zeta_novel (double **prob_zeta, int **edges, int M, double **probI, double *probA, double *probB)
{

  int m, s, t, i, j;
  
  for(m=1;m<=M;m++)
    {


      i = edges[0][m];
      j = edges[1][m];
      t = edges[2][m];
      s = edges[3][m];

      

      if(t == 0)
	{
	  prob_zeta[m][0] = probI[s][0] + probI[s][1] + probI[s][2];
	  prob_zeta[m][1] = probI[s][0] + probI[s][1];
	  prob_zeta[m][2] = probI[s][0] + probI[s][2];
	}

      if (t == 1){
	prob_zeta[m][0] = probA[s];
	prob_zeta[m][1] = probA[s];
	prob_zeta[m][2] = 0.0;
      }

      if (t == 2){
	prob_zeta[m][0] = probB[s];
	prob_zeta[m][1] = 0.0;
	prob_zeta[m][2] = probB[s];
      }

      //printf("-> %d %g %g %g\n", t,prob_zeta[m][0],prob_zeta[m][1], prob_zeta[m][2]);

    }

  //exit(0);
}



double single_iteration_novel (int M, int **matrix, int **edges, double **prob_zeta, int N, int **rev_bond, int **type_bond, int **bond_to_edge, double **probI, double *probA, double *probB, double p)
{
  int t, s, v, q, n, m, i, j;
  double err, max_err;


  double a, b, c;

  max_err = 0.0;


  compute_zeta_novel (prob_zeta, edges, M, probI, probA, probB);

  for(m=1;m<=M;m++)
    {

      i = edges[0][m];
      j = edges[1][m];
      t = edges[2][m];
      s = edges[3][m];
      
      a = b = c = 1.0;

      //printf("%d:\t",m);

      for(q=1;q<=matrix[0][m];q++)
	{

	  v = matrix[m][q];
	  
	  a *= (1.0 - prob_zeta[v][0]);
	  b *= (1.0 - prob_zeta[v][1]);
	  c *= (1.0 - prob_zeta[v][2]);

	  //printf("%d\t",v);

	}

      //printf(": %g %g %g\n",a,b,c);
	  

      if(t==0)
	{
	  err = probI[s][0] - p * (1.0 - b - c + a);
	  probI[s][0] = p * (1.0 - b - c + a);
	  
	  if (fabs(err)<fabs(probI[s][2] - p * (b - a))) err = probI[s][2] - p * (b - a);
	  probI[s][2] = p * (b - a);
	  if (fabs(err)<fabs(probI[s][1] - p * (c - a))) err = probI[s][1] - p * (c - a);
	  probI[s][1] = p * (c - a);

	  //if (fabs(err)<fabs(probI[s][1] - p * (b - a))) err = probI[s][1] - p * (b - a);
	  //probI[s][1] = p * (b - a);
	  //if (fabs(err)<fabs(probI[s][2] - p * (c - a))) err = probI[s][2] - p * (c - a);
	  //probI[s][2] = p * (c - a);
	}


      if(t==1)
	{
	  err = probA[s] - p * (1.0 - b - c + a);
	  probA[s] = p * (1.0 - b - c + a);
	}

      
      if(t==2)
	{
	  err = probB[s] - p * (1.0 - b - c + a);
	  probB[s] = p * (1.0 - b - c + a);
	}

      if(fabs(err)>max_err) max_err= fabs(err);



    }

  //exit(0);

  return max_err;
}




double compute_order (double **prob_zeta, int M, int **edges, int N, int **rev_bond, int **type_bond, int **bond_to_edge, double **probI, double *probA, double *probB, double p)
{


  compute_zeta_novel (prob_zeta, edges, M, probI, probA, probB);

  double order = 0.0;
  int m, i, j;
 

  double **ind_order = (double **)malloc((N+1)*sizeof(double *));
  for(i=1;i<=N;i++) 
    {
      ind_order[i] = (double *)malloc(3*sizeof(double));
      ind_order[i][0] = ind_order[i][1] = ind_order[i][2] = 1.0;
    }

  for(m=1;m<=M;m++)
    {
     
      i = edges[0][m];
      j = edges[1][m];
      

      ind_order[j][0] *= (1.0 - prob_zeta[m][0]);
      ind_order[j][1] *= (1.0 - prob_zeta[m][1]);
      ind_order[j][2] *= (1.0 - prob_zeta[m][2]);
	  
    }


  for(i=1;i<=N;i++) order += p* (1.0 - ind_order[i][1] - ind_order[i][2] + ind_order[i][0]);
  

  order /= (double)N;


  for(i=1;i<=N;i++) free(ind_order[i]);
  free(ind_order);



  return order;

}


